#include <memory>   // std::auto_ptr
#include <fstream>
#include <iostream>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/XMLUniDefs.hpp> // chLatin_*
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/validators/common/Grammar.hpp> // xercesc::Grammar
#include <xercesc/framework/Wrapper4InputSource.hpp>

#include <xsd/cxx/xml/dom/auto-ptr.hxx>
#include <xsd/cxx/xml/dom/bits/error-handler-proxy.hxx>
#include <xsd/cxx/xml/sax/std-input-source.hxx>

#include <xsd/cxx/tree/exceptions.hxx>
#include <xsd/cxx/tree/error-handler.hxx>

#include "citygml/CityGML.hxx"

using namespace std;

int
main (int argc, char* argv[])
{
  if (argc != 2)
  {
    cerr << "usage: " << argv[0] << " file.xml" << endl;
    return 1;
  }

  int r (0);

  // We need to initialize the Xerces-C++ runtime because we
  // are doing the XML-to-DOM parsing ourselves.
  //
  xercesc::XMLPlatformUtils::Initialize ();

  try
  {
    using namespace xercesc;
    namespace xml = xsd::cxx::xml;
    namespace tree = xsd::cxx::tree;

    const XMLCh ls_id [] = {chLatin_L, chLatin_S, chNull};

    // Get an implementation of the Load-Store (LS) interface.
    //
    DOMImplementation* impl (
      DOMImplementationRegistry::getDOMImplementation (ls_id));

    xml::dom::auto_ptr<DOMLSParser> parser (
      impl->createLSParser (DOMImplementationLS::MODE_SYNCHRONOUS, 0));

    DOMConfiguration* conf (parser->getDomConfig ());

    // Discard comment nodes in the document.
    //
    conf->setParameter (XMLUni::fgDOMComments, false);

    // Enable datatype normalization.
    //
    conf->setParameter (XMLUni::fgDOMDatatypeNormalization, true);

    // Do not create EntityReference nodes in the DOM tree. No
    // EntityReference nodes will be created, only the nodes
    // corresponding to their fully expanded substitution text
    // will be created.
    //
    conf->setParameter (XMLUni::fgDOMEntities, false);

    // Perform namespace processing.
    //
    conf->setParameter (XMLUni::fgDOMNamespaces, true);

    // Do not include ignorable whitespace in the DOM tree.
    //
    conf->setParameter (XMLUni::fgDOMElementContentWhitespace, false);

    // Enable/Disable validation.
    //
    conf->setParameter (XMLUni::fgDOMValidate, true);
    conf->setParameter (XMLUni::fgXercesSchema, true);
    conf->setParameter (XMLUni::fgXercesSchemaFullChecking, false);

    // Set error handler.
    //
    tree::error_handler<char> eh;
    xml::dom::bits::error_handler_proxy<char> ehp (eh);
    conf->setParameter (XMLUni::fgDOMErrorHandler, &ehp);

    // Initialize the schema cache.
    //
    parser->loadGrammar ("gml/gml.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/xAL.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/cityGMLBase.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/appearance.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/building.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/cityFurniture.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/cityObjectGroup.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/generics.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/landUse.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/relief.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/transportation.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/vegetation.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/waterBody.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/texturedSurface.xsd", Grammar::SchemaGrammarType, true);
    parser->loadGrammar ("citygml/CityGML.xsd", Grammar::SchemaGrammarType, true);


    conf->setParameter (XMLUni::fgXercesUseCachedGrammarInParse, true);

    // We will release the DOM document ourselves.
    //
    conf->setParameter (XMLUni::fgXercesUserAdoptsDOMDocument, true);

    // Parse XML documents.
    //
    ifstream ifs;
    ifs.exceptions (ifstream::badbit | ifstream::failbit);
    ifs.open (argv[1]);

    // Wrap the standard input stream.
    //
    xml::sax::std_input_source isrc (ifs, argv[1]);
    Wrapper4InputSource wrap (&isrc, false);

    // Parse XML to DOM.
    //
    xml_schema::dom::auto_ptr<DOMDocument> doc (parser->parse (&wrap));
    eh.throw_if_failed<tree::parsing<char> > ();

    // Parse DOM to the object model.
    //
    using namespace citygml;

    std::auto_ptr<CityModelType> city (CityModel (*doc));

    // Prepare namespace mapping and schema location information.
    //
    xml_schema::namespace_infomap map;

    map[""].name = "http://www.opengis.net/citygml/1.0";
    map["gml"].name = "http://www.opengis.net/gml";
    map["xlink"].name = "http://www.w3.org/1999/xlink";
    map["smil20lang"].name = "http://www.w3.org/2001/SMIL20/Language";
    map["smil20"].name = "http://www.w3.org/2001/SMIL20/";
    map["xAL"].name = "urn:oasis:names:tc:ciq:xsdschema:xAL:2.0";
    map["gen"].name = "http://www.opengis.net/citygml/generics/1.0";
    map["veg"].name = "http://www.opengis.net/citygml/vegetation/1.0";
    map["wtr"].name = "http://www.opengis.net/citygml/waterbody/1.0";
    map["trans"].name = "http://www.opengis.net/citygml/transportation/1.0";
    map["tex"].name = "http://www.opengis.net/citygml/texturedsurface/1.0";
    map["app"].name = "http://www.opengis.net/citygml/appearance/1.0";
    map["frn"].name = "http://www.opengis.net/citygml/cityfurniture/1.0";
    map["grp"].name = "http://www.opengis.net/citygml/cityobjectgroup/1.0";
    map["bldg"].name = "http://www.opengis.net/citygml/building/1.0";
    map["luse"].name = "http://www.opengis.net/citygml/landuse/1.0";
    map["dem"].name = "http://www.opengis.net/citygml/relief/1.0";
    map["xsi"].name = "http://www.w3.org/2001/XMLSchema-instance";

    // Write it out.
    //
    CityModel (std::cout,
               *city,
               map,
               "UTF-8",
               xml_schema::flags::dont_initialize);
  }
  catch (const xml_schema::exception& e)
  {
    cerr << e << endl;
    r = 1;
  }
  catch (const std::ios_base::failure&)
  {
    cerr << argv[1] << ": unable to open or read failure" << endl;
    r = 1;
  }

  xercesc::XMLPlatformUtils::Terminate ();
  return r;
}
